@@ -5,7 +5,7 @@ from paco.models.locations import get_parent_by_interface
 from functools import partial
 from zope.interface import implementer
 from zope.schema.fieldproperty import FieldProperty
-from paco.models.references import Reference, get_model_obj_from_ref
+from paco.models.references import get_model_obj_from_ref
 import json
 import hashlib
 import troposphere
@@ -14,6 +14,66 @@ import zope.schema
 import zope.schema.interfaces
 
 
+def match_allowed_paco_filenames(paco_home, sub_dirname, filename=None):
+    """
+    Return the Path to a Paco-named directory or file if it can match filename to one a lower-case
+    or capitalize case otherwise return None:
+
+    Directory:
+
+      * paco_home / {sub_dirname}
+      * paco_home / {Sub_dirname}
+
+    Lower-case sub_dirname:
+
+     * paco_home / {sub_dirname} / {filename}.yaml <-- lower-case
+     * paco_home / {sub_dirname} / {Filename}.yaml <-- capitalize
+     * paco_home / {sub_dirname} / {filename}.yml <-- .yml extension
+     * paco_home / {sub_dirname} / {Filename}.yml <-- .yml extension
+
+    Capitalized sub_dirname:
+
+     * paco_home / {Sub_dirname} / {filename}.yaml <-- lower-case
+     * paco_home / {Sub_dirname} / {Filename}.yaml <-- capitalize
+     * paco_home / {Sub_dirname} / {filename}.yml <-- .yml extension
+     * paco_home / {Sub_dirname} / {Filename}.yml <-- .yml extension
+
+    """
+    def filename_match(paco_home, sub_dirname, filename):
+        path = paco_home / sub_dirname / f'{filename}.yml'
+        if path.is_file():
+            return path
+        path = paco_home / sub_dirname / f'{filename}.yaml'
+        if path.is_file():
+            return path
+        filename = filename.capitalize()
+        path = paco_home / sub_dirname / f'{filename}.yml'
+        if path.is_file():
+            return path
+        path = paco_home / sub_dirname / f'{filename}.yaml'
+        if path.is_file():
+            return path
+
+    sub_dirname = sub_dirname.lower()
+    if filename == None:
+        path = paco_home / sub_dirname
+        if path.is_dir():
+            return path
+        sub_dirname = sub_dirname.capitalize()
+        path = paco_home / sub_dirname
+        if path.is_dir():
+            return path
+        else:
+            return None
+
+    filename = filename.lower()
+    path = filename_match(paco_home, sub_dirname, filename)
+    if path == None:
+        sub_dirname = sub_dirname.capitalize()
+        path = filename_match(paco_home, sub_dirname, filename)
+
+    return path
+
 def interface_seen(seen, iface):
     """Return True if interface already is seen.
     """
@@ -80,7 +140,7 @@ def marshall_fieldname_to_troposphere_value(obj, props, troposphere_name, field_
             else: return value
         else:
             # Shouldn't get here unless the cfn_mapping is incorrect
-            raise InvalidCFNMapping("cfn_mapping is incorrect for {} for {}?".format(mapping_field_name, interface))
+            raise InvalidCFNMapping(f"cfn_mapping is incorrect for {mapping_field_name} for {interface}?")
 
 def get_all_fields(obj):
     """
@@ -195,6 +255,7 @@ class Named(Parent):
     """
     name = FieldProperty(schemas.INamed["name"])
     title = FieldProperty(schemas.INamed["title"])
+    _stack_hooks = None
 
     def __init__(self, name, __parent__):
         super().__init__(__parent__)
@@ -208,6 +269,22 @@ class Named(Parent):
             return self.title
         return self.name
 
+    def add_stack_hooks(self, stack_hooks):
+        """
+        Add StackHooks to the Resource
+        """
+        # class attributes are shared, create an instance attribute
+        if self._stack_hooks == None:
+            self._stack_hooks = []
+
+        # If the Stack has already been initialized, add hooks to it
+        # ToDo: duck typing as paco.models can depend upon paco.stack - create Interface for IStack?
+        if hasattr(self, 'stack') and hasattr(self.stack, 'add_hooks'):
+            self.stack.add_hooks(stack_hooks)
+        # Before initialization the hooks are set as a private attribute and added during initialization
+        else:
+            self._stack_hooks.append(stack_hooks)
+
 
 @implementer(schemas.IName)
 class Name():
@@ -352,16 +429,19 @@ class Resource(Type, Named, Deployable, Regionalized, DNSEnablable):
     order = FieldProperty(schemas.IResource['order'])
     change_protected = FieldProperty(schemas.IResource['change_protected'])
 
-    def resolve_ref(self, ref):
-        return self.resolve_ref_obj.resolve_ref(ref)
-
     def get_account(self):
         """
         Return the Account object that this resource is provisioned to
         """
-        env_reg = get_parent_by_interface(self, schemas.IEnvironmentRegion)
+        region = get_parent_by_interface(self, schemas.IRegionContainer)
         project = get_parent_by_interface(self, schemas.IProject)
-        return get_model_obj_from_ref(env_reg.network.aws_account, project)
+        # NetEnv accounts
+        if schemas.IEnvironmentRegion.providedBy(region):
+            return get_model_obj_from_ref(region.network.aws_account, project)
+        # Service accounts
+        else:
+            account_cont = get_parent_by_interface(self, schemas.IAccountContainer)
+            return project.accounts[account_cont.name]
 
     # Resource Name methods
 
@@ -535,6 +615,15 @@ class Resource(Type, Named, Deployable, Regionalized, DNSEnablable):
             camel_case=camel_case
         )
 
+class Stack():
+    "Unit of provisionable cloud resources"
+
+class HasStack():
+    "Resource that has a Stack associated with it"
+    stack = None
+
+    def resolve_ref(self, ref):
+        return self.stack
 
 @implementer(schemas.IApplicationResource)
 class ApplicationResource(Resource):
