@@ -2,23 +2,20 @@
 All things Resources.
 """
 
-import json
 import troposphere.apigateway
 import troposphere.route53
-from paco.models.base import Parent, Named, CFNExport, Deployable, Regionalized, Resource, ApplicationResource, AccountRegions
-from paco.models.metrics import Monitorable
+from paco.models.base import Enablable, Parent, Named, CFNExport, Deployable, HasStack, Resource, ApplicationResource, AccountRegions
+from paco.models.metrics import SNSTopics
 from paco.models import references
 from paco.models import schemas
 from zope.interface import implementer
 from zope.schema.fieldproperty import FieldProperty
 from paco.models import loader
 from paco.models.locations import get_parent_by_interface
+from paco.models.applications import DNS
 from paco.models.references import Reference
 from paco.models import references
 
-@implementer(schemas.IGlobalResources)
-class GlobalResources(Named, dict):
-    "Global Resources"
 
 @implementer(schemas.ITopics)
 class Topics(Named, dict):
@@ -35,10 +32,21 @@ class SNS(Named, dict):
     def __init__(self, name, parent):
         super().__init__(name, parent)
         self.computed = Computed('computed', self)
+        self.topics = Topics('topics', self)
 
     def resolve_ref(self, ref):
         return self.resolve_ref_obj.resolve_ref(ref)
 
+@implementer(schemas.IApiGatewayBasePathMapping)
+class ApiGatewayBasePathMapping(Parent):
+    base_path = FieldProperty(schemas.IApiGatewayBasePathMapping['base_path'])
+    stage = FieldProperty(schemas.IApiGatewayBasePathMapping['stage'])
+
+@implementer(schemas.IApiGatewayDNS)
+class ApiGatewayDNS(DNS):
+    base_path_mappings = FieldProperty(schemas.IApiGatewayDNS['base_path_mappings'])
+    ssl_certificate = FieldProperty(schemas.IApiGatewayDNS['ssl_certificate'])
+
 @implementer(schemas.IApiGatewayMethods)
 class ApiGatewayMethods(Named, dict):
     pass
@@ -62,11 +70,13 @@ class ApiGatewayMethodIntegrationResponse(CFNExport):
 
 @implementer(schemas.IApiGatewayMethodIntegration)
 class ApiGatewayMethodIntegration(Parent, CFNExport):
-    integration_responses = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_responses'])
-    request_parameters = FieldProperty(schemas.IApiGatewayMethodIntegration['request_parameters'])
     integration_http_method = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_http_method'])
-    integration_type = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_type'])
     integration_lambda = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_lambda'])
+    integration_responses = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_responses'])
+    integration_type = FieldProperty(schemas.IApiGatewayMethodIntegration['integration_type'])
+    pass_through_behavior = FieldProperty(schemas.IApiGatewayMethodIntegration['pass_through_behavior'])
+    request_parameters = FieldProperty(schemas.IApiGatewayMethodIntegration['request_parameters'])
+    request_templates = FieldProperty(schemas.IApiGatewayMethodIntegration['request_templates'])
     uri = FieldProperty(schemas.IApiGatewayMethodIntegration['uri'])
 
     @property
@@ -83,8 +93,8 @@ class ApiGatewayMethodIntegration(Parent, CFNExport):
         #"ConnectionId": (basestring, False),
         #"ConnectionType": (basestring, False),
         #"ContentHandling": (basestring, False),
-        #"PassthroughBehavior": (basestring, False),
-        #"RequestTemplates": (dict, False),
+        "PassthroughBehavior": 'pass_through_behavior',
+        "RequestTemplates": 'request_templates',
         #"TimeoutInMillis": (integer_range(50, 29000), False),
         "IntegrationResponses": 'integration_responses_cfn',
         "IntegrationHttpMethod": 'integration_http_method',
@@ -94,6 +104,15 @@ class ApiGatewayMethodIntegration(Parent, CFNExport):
         #"Uri": computed in the template,
     }
 
+@implementer(schemas.IApiGatewayCognitoAuthorizers)
+class ApiGatewayCognitoAuthorizers(Named, dict):
+    pass
+
+@implementer(schemas.IApiGatewayCognitoAuthorizer)
+class ApiGatewayCognitoAuthorizer(Named):
+    identity_source = FieldProperty(schemas.IApiGatewayCognitoAuthorizer['identity_source'])
+    user_pools = FieldProperty(schemas.IApiGatewayCognitoAuthorizer['user_pools'])
+
 @implementer(schemas.IApiGatewayMethodMethodResponseModel)
 class ApiGatewayMethodMethodResponseModel():
     content_type = FieldProperty(schemas.IApiGatewayMethodMethodResponseModel['content_type'])
@@ -103,11 +122,14 @@ class ApiGatewayMethodMethodResponseModel():
 class ApiGatewayMethodMethodResponse():
     status_code = FieldProperty(schemas.IApiGatewayMethodMethodResponse['status_code'])
     response_models = FieldProperty(schemas.IApiGatewayMethodMethodResponse['response_models'])
+    response_parameters = FieldProperty(schemas.IApiGatewayMethodMethodResponse['response_parameters'])
 
 @implementer(schemas.IApiGatewayMethod)
 class ApiGatewayMethod(Resource):
     type = "ApiGatewayMethod"
-    resource_id = FieldProperty(schemas.IApiGatewayMethod['resource_id'])
+    authorization_type = FieldProperty(schemas.IApiGatewayMethod['authorization_type'])
+    authorizer = FieldProperty(schemas.IApiGatewayMethod['authorizer'])
+    resource_name = FieldProperty(schemas.IApiGatewayMethod['resource_name'])
     http_method = FieldProperty(schemas.IApiGatewayMethod['http_method'])
     request_parameters = FieldProperty(schemas.IApiGatewayMethod['request_parameters'])
     method_responses = FieldProperty(schemas.IApiGatewayMethod['method_responses'])
@@ -117,6 +139,14 @@ class ApiGatewayMethod(Resource):
     def integration_cfn(self):
         return self.integration.cfn_export_dict
 
+    def get_resource(self):
+        resource = None
+        node = get_parent_by_interface(self, schemas.IApiGatewayRestApi).resources
+        for part in self.resource_name.split('.'):
+            resource = node[part]
+            node = resource.child_resources
+        return resource
+
     troposphere_props = troposphere.apigateway.Method.props
     cfn_mapping = {
         #"ApiKeyRequired": (bool, False),
@@ -158,16 +188,27 @@ class ApiGatewayResources(Named, dict):
     pass
 
 @implementer(schemas.IApiGatewayResource)
-class ApiGatewayResource(Resource):
+class ApiGatewayResource(Named, dict):
     type = "ApiGatewayResource"
-    parent_id = FieldProperty(schemas.IApiGatewayResource['parent_id'])
     path_part = FieldProperty(schemas.IApiGatewayResource['path_part'])
+    enable_cors = FieldProperty(schemas.IApiGatewayResource['enable_cors'])
+    # child_resources is not a FieldProperty as it's set during recursion
+
+    @property
+    def nested_name(self):
+        parent = self
+        parts = []
+        while not schemas.IApiGatewayRestApi.providedBy(parent):
+            parts.append(parent.name)
+            parent = parent.__parent__.__parent__
+        parts.reverse()
+        return '.'.join(parts)
 
     troposphere_props = troposphere.apigateway.Resource.props
     cfn_mapping = {
-        # ParentId needs to be computed from the CFNTemplate to inject the Resource's Logical Id into it
+        # ParentId: computed in template
         "PathPart": 'path_part',
-        # "RestApiId": 'rest_api_id' - Computed to CFNTemplate Logical Id
+        # "RestApiId": computed in template
     }
 
 @implementer(schemas.IApiGatewayStages)
@@ -187,15 +228,16 @@ class ApiGatewayStage(Resource):
     }
 
 @implementer(schemas.IApiGatewayRestApi)
-class ApiGatewayRestApi(ApplicationResource):
-    title = "API Gateway REST API"
+class ApiGatewayRestApi(ApplicationResource, HasStack):
     type = "ApiGatewayRestApi"
     api_key_source_type = FieldProperty(schemas.IApiGatewayRestApi['api_key_source_type'])
     binary_media_types = FieldProperty(schemas.IApiGatewayRestApi['binary_media_types'])
     body_file_location = FieldProperty(schemas.IApiGatewayRestApi['body_file_location'])
     body_s3_location = FieldProperty(schemas.IApiGatewayRestApi['body_s3_location'])
     clone_from = FieldProperty(schemas.IApiGatewayRestApi['clone_from'])
+    cognito_authorizers = FieldProperty(schemas.IApiGatewayRestApi['cognito_authorizers'])
     description = FieldProperty(schemas.IApiGatewayRestApi['description'])
+    dns = FieldProperty(schemas.IApiGatewayRestApi['dns'])
     endpoint_configuration = FieldProperty(schemas.IApiGatewayRestApi['endpoint_configuration'])
     fail_on_warnings = FieldProperty(schemas.IApiGatewayRestApi['fail_on_warnings'])
     minimum_compression_size = FieldProperty(schemas.IApiGatewayRestApi['minimum_compression_size'])
@@ -235,7 +277,7 @@ class ApiGatewayRestApi(ApplicationResource):
         "BinaryMediaTypes": 'binary_media_types',
         "CloneFrom": 'clone_from',
         "Description": 'description',
-        "Name": 'name',
+        # "Name": supplied by CloudFormation
         "Body": 'body',
         "FailOnWarnings": 'fail_on_warnings',
         "EndpointConfiguration": ("endpoint_configuration", "endpoint_configuration_cfn"),
@@ -342,7 +384,6 @@ class Route53HostedZone(Named, Deployable):
 @implementer(schemas.IRoute53Resource)
 class Route53Resource(Named):
     name = 'route53'
-    title = 'Route53'
     hosted_zones = FieldProperty(schemas.IRoute53Resource["hosted_zones"])
 
     def __init__(self, name, parent, config_dict):
@@ -460,7 +501,7 @@ class CodeCommitRepository(Named, Deployable):
 
     def __init__(self, name, parent):
         super().__init__(name, parent)
-        self.user = CodeCommitUsers('users', self)
+        self.users = CodeCommitUsers('users', self)
 
 @implementer(schemas.ICodeCommitRepositoryGroup)
 class CodeCommitRepositoryGroup(Named, dict):
@@ -505,6 +546,16 @@ class CodeCommit(Named, dict):
     def resolve_ref(self, ref):
         return self.resolve_ref_obj.resolve_ref(ref)
 
+    def repo_list(self):
+        "List of all repositories"
+        out = []
+        for group in self.values():
+            for repo in group.values():
+                repo.repository_group = group
+                out.append(repo)
+        return out
+
+
 @implementer(schemas.IConfig)
 class Config(Resource):
     delivery_frequency = FieldProperty(schemas.IConfig["delivery_frequency"])
@@ -576,7 +627,7 @@ class CloudTrailResource(Named):
 
 
 @implementer(schemas.IIAMUserProgrammaticAccess)
-class IAMUserProgrammaticAccess(Deployable):
+class IAMUserProgrammaticAccess(Enablable):
     access_key_1_version = FieldProperty(schemas.IIAMUserProgrammaticAccess['access_key_1_version'])
     access_key_2_version = FieldProperty(schemas.IIAMUserProgrammaticAccess['access_key_2_version'])
 
@@ -607,6 +658,17 @@ class IAMUserPermissionCodeBuildResource(Parent):
     permission = FieldProperty(schemas.IIAMUserPermissionCodeBuildResource['permission'])
     console_access_enabled = FieldProperty(schemas.IIAMUserPermissionCodeBuildResource['console_access_enabled'])
 
+@implementer(schemas.IIAMUserPermissionDeploymentPipelines)
+class IAMUserPermissionDeploymentPipelines(IAMUserPermission):
+    resources = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelines['resources'])
+    accounts = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelines['accounts'])
+
+@implementer(schemas.IIAMUserPermissionDeploymentPipelineResource)
+class IAMUserPermissionDeploymentPipelineResource(Parent):
+    pipeline = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['pipeline'])
+    permission = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['permission'])
+    console_access_enabled = FieldProperty(schemas.IIAMUserPermissionDeploymentPipelineResource['console_access_enabled'])
+
 @implementer(schemas.IIAMUserPermissionCodeBuild)
 class IAMUserPermissionCodeBuild(IAMUserPermission):
     resources = FieldProperty(schemas.IIAMUserPermissionCodeBuild['resources'])
@@ -636,6 +698,10 @@ class IAMUser(Named, Deployable):
     permissions = FieldProperty(schemas.IIAMUser['permissions'])
     account_whitelist = FieldProperty(schemas.IIAMUser['account_whitelist'])
 
+    def __init__(self, name, __parent__):
+        super().__init__(name, __parent__)
+        self.permissions = IAMUserPermissions('permissions', self)
+
 @implementer(schemas.IIAMUsers)
 class IAMUsers(Named, dict):
     pass
@@ -674,9 +740,25 @@ class SSMDocument(Resource):
 @implementer(schemas.ISSMResource)
 class SSMResource(Named):
     name = 'ssm'
-    title = 'SSM'
     ssm_documents = FieldProperty(schemas.ISSMResource["ssm_documents"])
 
     def __init__(self, name, parent):
         super().__init__(name, parent)
         self.ssm_documents = SSMDocuments('ssm_documents', self)
+
+@implementer(schemas.IGlobalResources)
+class GlobalResources(Named, dict):
+    "Global Resources"
+
+    def __init__(self, name, __parent__):
+        super().__init__(name, __parent__)
+        self['cloudtrail'] = CloudTrailResource('cloudtrail', self)
+        self.cloudtrail = self['cloudtrail']
+        self['codecommit'] = CodeCommit('codecommit', self)
+        self.codecommit = self['codecommit']
+        self['sns'] = SNS('sns', self)
+        self.sns = self['sns']
+        self['snstopics'] = SNSTopics('snstopics', self)
+        self.snstopics = self['snstopics']
+        self['iam'] = IAMResource('iam', self)
+        self.iam = self['iam']
